////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2009, RoboRealm, all rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are NOT permitted.
//
// Refer to http://www.roborealm.com/registration/Source_Code_Agreement.html
// for license information.
////////////////////////////////////////////////////////////////////////////

import java.io.*;
import java.net.*;
import java.awt.*;
import java.awt.image.*;
import java.applet.Applet;
import java.util.*;

public class RoboRealm extends Applet implements Runnable
{
  private boolean running=true;
  private Image image;

  private URL url;
  private DataInputStream dis;
  private InputStream is;

  private Socket source;
  private InputStream in;
  private OutputStream out;

  private boolean connected=false;

  public long bytesRead=0;
  private int port;
  private String authorization = null;

  // used to time FPS
  private long timeArray[] = new long[32];
  private int timeIndex = 0;
  private long baudArray[] = new long[32];

  private int timeout = 1200000;

  private int frameNum;
  private int size;
  private long lastMotion;

  private int width, height;

  Thread animationThread = null;

  private Font stoppedFont = new Font("Helvetica", Font.BOLD, 20);

  public RoboRealm()
  {
  }

  public void init()
  {
    //addNotify();
    //setLayout(null);
    //setBackground(Color.black);

    lastMotion=System.currentTimeMillis();

    image = createImage(10,10);

    try
    {
      String s;
      if ((s=getParameter("timeout"))!=null)
        timeout = Integer.parseInt(s)*1000;

      authorization = getParameter("authorization");
    }
    catch (Exception e)
    {
      System.out.println(e.getMessage());
    };

  }

  public void reshape(int x, int y, int w, int h)
  {
    super.reshape(x, y, w, h);

    width = w;
    height = h;
  }

  public void calculateFPS()
  {
    long j = System.currentTimeMillis();
    long tmp, tmp1, tmp2;

    if (timeout!=0)
    {
      if ((System.currentTimeMillis()-lastMotion)>timeout)
      {
        running=false;
        close();
        return;
      }
    }

    if ((timeIndex%8)==0)
    {
      tmp = ((j - timeArray[timeIndex])/10L);

//System.out.println(tmp);

      if (tmp>0)
      {
        tmp1 = 320000/tmp;
        tmp2 = (int)(((float)((bytesRead - baudArray[timeIndex])*100)/tmp)/10.24);
        // keep the /100 so we get two decimal places in the printout
        getAppletContext().showStatus(new StringBuffer(String.valueOf(tmp1/100.0)).append(" fps (").append((tmp2/100.0)).append(" Kps)  #").append(frameNum).toString());
      }
    }

    baudArray[timeIndex] = bytesRead;
    timeArray[timeIndex] = j;

    timeIndex++;
    if (timeIndex>=32) timeIndex=0;
  }

  public final void start()
  {
    //super.start();

    //reshape(0, 0, width, height);

    frameNum=0;
    running = true;

    if (animationThread == null)
    {
      animationThread = new Thread(this);
      animationThread.start();
    }
  }

  public final void stop()
  {
    //super.stop();

    if (animationThread!=null)
    {
      animationThread.stop();
      animationThread = null;
    }
    running=false;
    close();
  }

  /********************************************************************************************
  Display
  ********************************************************************************************/

  public void update(Graphics g)
  {
    paint(g);
  }

  public void paint(Graphics g)
  {
    g.drawImage(image, 0, 0, this);

    if (!running)
    {
       String txt="Video Stopped";
       g.setColor(Color.black);
       g.setFont(stoppedFont);
       g.drawString(txt, (width>>1)-65, height-10);
       g.setColor(Color.red);
       g.drawString(txt, (width>>1)-65, height-10);
       getAppletContext().showStatus(txt+" ... click to continue.");
    }
  }

  public void run()
  {
    Graphics g=getGraphics();

    byte pixels[] =new byte[320*240*3];

    while (running)
    {
      while (running)
      {
        String txt = readLine();
        if (txt==null) continue;
        if (txt.equals("--4532345678643234567888543232245\r\n")) break;
      }

      int len=0;
      while (true)
      {
        String txt = readLine();
        if (txt==null) break;
        if (txt.equals("\r\n")) break;
        if (txt.startsWith("Content-Length: "))
        {
          len = Integer.parseInt(txt.substring(16).trim());
        }
      }

      if (len==0)
      {
        close();
        break;
      }

//System.out.println(len);

      read(pixels, len);

      image.flush();
      image=getToolkit().createImage(pixels);

      g.drawImage(image, 0, 0, this);
      getToolkit().sync();
      frameNum++;
      calculateFPS();
    }

    animationThread = null;
  }

  private int getPort()
  {
    String s;

    try
    {
      if ((s=getParameter("port"))!=null)
        return Integer.parseInt(s);
      else
        return -1;
    }
    catch (Exception e)
    {
      System.out.println(e.getMessage());
    };

    return -1;
  }

  private boolean open()
  {
    int port;
    String token;
    String s;
    int i;

    for (i=0;i<32;i++)
    {
      timeArray[i]=0;
      baudArray[i]=0;
    }

    try
    {
      url=getDocumentBase();
      if (url.getHost().length()==0)
      {
        try
        {
          url=new URL("http://127.0.0.1:8080");
        }
        catch (Exception e){};
      }

//url=new URL("http://127.0.0.1:8080");

      port = getPort();
      if (port==-1)
        port=url.getPort();
      if (port==-1)
        port=8080;

      source = new Socket(url.getHost(),port);
      dis = new DataInputStream(source.getInputStream());
      out  = source.getOutputStream();

      if (authorization!=null)
        out.write(("GET /mjpeg.cgi HTTP 1.1\r\nAuthorization: Basic "+authorization+"\r\n\r\n").getBytes());
      else
        out.write("GET /mjpeg.cgi HTTP 1.1\r\n\r\n".getBytes());

      connected=true;
    }
    catch (IOException e)
    {
      connected=false;
      return false;
    };

    return true;
  }

  public void close()
  {
    connected=false;
    try
    {
      dis.close();
    }
    catch (Exception e)
    {
    };
    System.out.println("Lost Connection! Attempting to reconnect...");
  }

  public int read(byte data[], int len)
  {
    int num,i=0;

    try
    {
      while (len>0)
      {
        if ((num = dis.read(data,i,len))<0)
        {
          close();
          return -1;
        }
        len-=num;
        i+=num;
        bytesRead+=num;
      }
    }
    catch (Exception e)
    {
      close();
      return -1;
    }
    return i;
  }

  private String readLine()
  {
    if (!connected)
    {
      if (!open())
      {
        getAppletContext().showStatus("Connection failure!");
        return null;
      }

      getAppletContext().showStatus("Streaming Video...");
    }

    StringBuffer sb = new StringBuffer();
    while (true)
    {
      try
      {
        int c = dis.read();
        if (c==-1)
        {
          connected=false;
          return null;
        }
        if (c=='\n')
        {
          sb.append((char)c);
          return sb.toString();
        }
        sb.append((char)c);
      }
      catch (Exception e)
      {
        return null;
      }
    }
  }

  public boolean handleEvent(Event e)
  {
    if (e.id == Event.MOUSE_MOVE)
    {
      lastMotion = System.currentTimeMillis();
      if (!running)
      {
        start();
      }
      return false;
    }

    if (e.id == Event.MOUSE_DOWN)
    {
      if (e.target==this)
      {
        if (!running)
        {
          start();
        }
        return true;
      }
      return false;
    }
    return true;
  }
}
